/* NRCRandom.h
 * Random Number Generation functions from Numerical Recipes in C
 */

#ifndef _NRCRANDOM_H
#define _NRCRANDOM_H

#ifdef __cplusplus
extern "C" {
#endif

/** Returns as a floating-point number an integer value that is a random
	deviate drawn from the binomial distribution of n trials each of
	probability pp, using ran1(idum) as a source of uniform random deviates.
	See Numerical Recipes in C, Section 7.3, p.295.
*/
double bnldev(double pp, int n, long *idum);
	
/** Returns a normally distributed deviate with zero mean and unit variance,
	using ran1(idum) as the source of uniform deviates.
	See Numerical Recipes in C, Section 7.2, p.289.
*/
double gasdev(long *idum);
	
/** "Minimal" random number generator of Park and Miller with Bays-Durham
	shuffle and added safeguards. Returns a uniform random deviate between
	0.0 and 1.0 (exclusive of the endpoint values). Call with idum a negative
	integer to initialize; thereafter, do not alter idum between successive
	deviates in a sequence. RNMX should approximate the largest floating
	value that is less than 1.
	See Numerical Recipes in C, Section 7.1, p.280.
*/
double ran1(long *idum);

/** Calculate the natural logarithm of the gamma function of xx.
	See Numerical Recipes in C, Section 6.1, p.214.
*/
double gammln(double xx);

#ifdef __cplusplus
}
#endif

#endif
