// Organism.h

#ifndef __Organism_h__
#define __Organism_h__

#include "Gene.h"
#include "EnvVector.h"
#include "OrgID.h"
#include <vector>

class Simulator;

class Organism
{
public:
	Organism( int timeOfBirth = 0, bool bRandomCreate = true, Simulator * pSim = NULL );
	Organism( Organism * pParent, int timeOfBirth, Simulator * pSim, bool bMutate = true );
	~Organism();

	void randomCreate( Simulator * pSim = NULL );
	void calculateFitness( EnvVector & env );
	float getProbDeath() const;
	float getProbReproduction() const;
	int getNumGenes() const;
	int getSpecificity() const;
	OrgID & getID();
	int getTimeOfBirth() const;
	int getNumOffspring() const;
	void recordNewOffspring();

	Gene & getEnvDisturbance();
	
	friend ostream& operator << (ostream& os, Organism& id);

private:
	// methods
	void updateSpecificity();
	void updatePhenotype();
	float hammingDistance( EnvVector & env );

	// data members
	int m_ReqStartPos;
	vector<Gene> m_Genome;
	Gene m_EnvDisturbance;

	EnvVector m_Phenotype;

	float m_Fitness;
	float m_ProbDeath;
	float m_ProbReproduction;
	int m_Specificity;
	float m_SpecificityFraction;

	OrgID m_ID;
	OrgID m_ParentID;
	int m_TimeOfBirth;
	int m_NumOffspring;
	
	bool m_bGenomeRecorded;
	
};


inline Organism::~Organism()
{
}


inline Gene & Organism::getEnvDisturbance()
{
	return m_EnvDisturbance;
}


inline float Organism::getProbDeath() const
{
	return m_ProbDeath;
}


inline float Organism::getProbReproduction() const
{
	return m_ProbReproduction;
}


inline int Organism::getNumGenes() const
{
	return (int)(m_Genome.size());
}


inline int Organism::getSpecificity() const
{
	return m_Specificity;
}


inline OrgID & Organism::getID()
{
	return m_ID;
}


inline int Organism::getTimeOfBirth() const
{
	return m_TimeOfBirth;
}


inline int Organism::getNumOffspring() const
{
	return m_NumOffspring;
}


#endif
