// Patch.h

#ifndef __Patch_h__
#define __Patch_h__

#include "EnvVector.h"
#include <vector>

class Organism;
class Simulator;

class Patch
{
	
public:
	
	Patch();
	~Patch();
	
	void initialise(Simulator * pSim, int x, int y);

	void addOrganism( Organism * pOrg );
	
	int getNumOrganisms();

	vector<Organism *> & getOrganisms();
	
	/** Update the local environment of this patch according to the organisms present at this
	and neighbouring patches */
	void updateEnvironment();

	/** Update the fitness scores and associated measures (probability of reproduction and
	probability of death) for each organism in the patch */
	void updateOrganisms();

	void killOrganismsStochastic();		// consider killing each organism based on its death prob

	void killOrganism(); 				// kill the least fit organism

	void killOrganismRouletteWheel();	// kill one organism using roulette wheel selection according to death probs

	void reproduceOrganisms( int iteration );

	Patch & getNeighbour(int ix, int iy);

	void test();
	
	friend ostream& operator << (ostream& os, const Patch& id);
	

private:
	// methods

	/** Modify the given environment vector according to the organisms currently in this patch.
	N.B. After this method has been called, *pEnv may contain values > 1. It therefore needs to
	be renormalised before being used again. */
	int addEnvDisturbances( EnvVector * pEnv );

	// data members

	vector<Organism*> m_Organisms;
	EnvVector m_LocalAbioticEnv;
	EnvVector m_CurrentModifiedEnv;
	Simulator * m_pSim;
	int m_X;
	int m_Y;

};


inline int Patch::getNumOrganisms()
{
	return m_Organisms.size();
}


inline void Patch::initialise(Simulator * pSim, int x, int y)
{
	m_pSim = pSim;
	m_X = x;
	m_Y = y;
}


inline vector<Organism *> & Patch::getOrganisms()
{
	return m_Organisms;
}


#endif
