// EnvVector.cc

#include "EnvVector.h"
#include "Gene.h"
#include "Simulator.h"
#include <iostream>
#include <iomanip>
#include <cassert>


void EnvVector::modify( const Gene * pGene )
{
	EnvVector modifier;
	int startPos;
	
	getModifier( pGene, &modifier, &startPos );

	switch (pGene->m_Operator)
	{
		case  0: operatorMult( startPos, modifier ); break;
		case  1: operatorUnion( startPos, modifier ); break;
		case  2: operatorIntersection( startPos, modifier ); break;
		case  3: operatorCompUnion( startPos, modifier ); break;
		case  4: operatorCompIntersection( startPos, modifier ); break;
		case  5: operatorAvg( startPos, modifier ); break;
		case  6: operatorCompAvg( startPos, modifier ); break;
		case  7: operatorAUnionCompB( startPos, modifier ); break;
		case  8: operatorAIntersectionCompB( startPos, modifier ); break;
		case  9: operatorCompAUnionB( startPos, modifier ); break;
		case 10: operatorCompAIntersectionB( startPos, modifier ); break;
		case 11: operatorCompMult( startPos, modifier ); break;
		case 12: operatorCompAUnionMult( startPos, modifier ); break;
		case 13: operatorCompAIntersectionMult( startPos, modifier ); break;
		case 14: operatorCompBUnionMult( startPos, modifier ); break;
		case 15: operatorCompBIntersectionMult( startPos, modifier ); break;
		default:
		{
			cerr << "In method EnvVector::modify. Unexpected operator " << pGene->m_Operator << "!" << endl;
			exit( EXIT_FAILURE );
		}			
	}
}


void EnvVector::add( const Gene * pGene )
{
	EnvVector modifier;
	int startPos;
	
	getModifier( pGene, &modifier, &startPos );
	
	vector<float> & vec = modifier.getValues();

	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] += vec[i];
		//assert(vec[i] <= 1.0f);
	}
}


void EnvVector::getModifier( const Gene * pGene, EnvVector * pModifier, int * pStartPos )
{
	static const float fPARTITION_SIZE = (float)Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE / (float)Gene::NUM_PARTITIONS;
	
	int vecSize = (int) ((fPARTITION_SIZE + Simulator::FLOATING_POINT_ACCURACY) * (float)(1 + (pGene->m_Spread * 2)));
	
	vector<float> & vec = pModifier->m_Values;
	vec.clear();
	for (int i = 0; i < vecSize; ++i)
		vec.push_back( 0.0f );
	
	int modifierTotalLength = vec.size();
	int modifierSingleRepeatLength = (int)(((float)modifierTotalLength / (float)(pGene->m_Repeat+1)) + Simulator::FLOATING_POINT_ACCURACY );
	if (modifierSingleRepeatLength < 2)
	{
		cerr << "In EnvVector::modify, length of single instance of pattern is too small (" << modifierSingleRepeatLength << ")" << endl;
		exit( EXIT_FAILURE );
	}
	
	int modifierSingleRepeatLengthFinal = modifierSingleRepeatLength + (modifierTotalLength - (modifierSingleRepeatLength * (pGene->m_Repeat+1)));
	//cout << "modifierTotalLength=" << modifierTotalLength << ", repeat=" << pGene->m_Repeat << "(+1), modifierSingleRepeatLength=" << modifierSingleRepeatLength
	//	<< ", modifierSingleRepeatLengthFinal=" << modifierSingleRepeatLengthFinal << endl;
	
	bool bInvert = false;
	switch (pGene->m_Function)
	{
		case  0: generateFunction0( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); break;
		case  1: generateFunction0( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); bInvert = true; break;
		case  2: generateFunction1( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); break;
		case  3: generateFunction1( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); bInvert = true; break;
		case  4: generateFunction2( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); break;
		case  5: generateFunction2( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); bInvert = true; break;
		case  6: generateFunction3( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); break;
		case  7: generateFunction3( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); bInvert = true; break;
		case  8: generateFunction4( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); break;
		case  9: generateFunction4( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); bInvert = true; break;
		case 10: generateFunction5( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); break;
		case 11: generateFunction5( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); bInvert = true; break;
		case 12: generateFunction6( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); break;
		case 13: generateFunction6( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); bInvert = true; break;
		case 14: generateFunction7( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); break;
		case 15: generateFunction7( vec, pGene->m_Repeat, modifierSingleRepeatLength, modifierSingleRepeatLengthFinal ); bInvert = true; break;
		default:
		{
			cerr << "In method EnvVector::modify. Unexpected function " << pGene->m_Function << "!" << endl;
			exit( EXIT_FAILURE );
		}
	}
	
	// invert if necessary
	if (bInvert)
		for (int i = 0; i < modifierTotalLength; ++i)
			vec[i] = 1.0f - vec[i];
		
	// and set start pos
	*pStartPos = (int) ( fPARTITION_SIZE * (float)(pGene->m_Partition - pGene->m_Spread) + Simulator::FLOATING_POINT_ACCURACY );
	if (*pStartPos < 0)
		*pStartPos += Simulator::LOCAL_ENVIRONMENT_VECTOR_SIZE;		
}


void EnvVector::divide( float divisor )
{
	vector<float>::iterator env_it  = m_Values.begin();
	vector<float>::iterator env_end = m_Values.end();
	while (env_it != env_end)
	{
		//cout << (*env_it) << "/" << divisor;
		(*env_it) /= divisor;
		//cout << "=" << (*env_it) << endl;
		assert( *env_it <= 1.0f );
		++env_it;
	}
}


void EnvVector::generateFunction0( vector<float> & vec, int repeat, int patternLength, int patternLengthFinal )
{
	float l = (float)(patternLength-1);
	int lp1 = patternLength;
	
	for (int r = 0; r <= repeat; ++r)
	{
		if (r == repeat)
		{
			l = (float)(patternLengthFinal-1);
			lp1 = patternLengthFinal;
		}
		
		for (int i = 0; i < lp1; ++i)
			vec[r*patternLength+i] = mybounded( ((float)i)/l );
	}
}


void EnvVector::generateFunction1( vector<float> & vec, int repeat, int patternLength, int patternLengthFinal )
{
	int i, r;
	float l = (float)(patternLength-1);
	int lp1 = patternLength;
	int d = (lp1-1)/2;

	for (r = 0; r <= repeat; ++r)
	{
		if (r == repeat)
		{
			l = (float)(patternLengthFinal-1);
			lp1 = patternLengthFinal;
			d = (lp1-1)/2;
		}

		for (i = 0; i < d; ++i)
			vec[r*patternLength+i] = mybounded( ((float)(2*i))/l );

		for (; i < lp1; ++i)
			vec[r*patternLength+i] = mybounded( 2.0f - ((float)(2*i))/l );
	}
}


void EnvVector::generateFunction2( vector<float> & vec, int repeat, int patternLength, int patternLengthFinal )
{
	int i, r;
	int lp1 = patternLength;
	int d = (lp1-1)/2;

	for (r = 0; r <= repeat; ++r)
	{
		if (r == repeat)
		{
			lp1 = patternLengthFinal;
			d = (lp1-1)/2;
		}
		
		for (i = 0; i < d; ++i)
			vec[r*patternLength+i] = 0.25f;

		for (; i < lp1; ++i)
			vec[r*patternLength+i] = 0.75f;
	}
}


void EnvVector::generateFunction3( vector<float> & vec, int repeat, int patternLength, int patternLengthFinal )
{
	int i, r;
	float l = (float)(patternLength-1);
	int lp1 = patternLength;
	int d1 = (lp1-1)/4;
	int d2 = (3*(lp1-1))/4;

	for (r = 0; r <= repeat; ++r)
	{
		if (r == repeat)
		{
			l = (float)(patternLengthFinal-1);
			lp1 = patternLengthFinal;
			d1 = (lp1-1)/4;
			d2 = (3*(lp1-1))/4;			
		}
		
		for (i = 0; i < d1; ++i)
			vec[r*patternLength+i] = mybounded( ((float)(4*i))/l );

		for (; i < d2; ++i)
			vec[r*patternLength+i] = 1.0f;		

		for (; i < lp1; ++i)
			vec[r*patternLength+i] = mybounded( 4.0f - ((float)(4*i))/l );
	}
}


void EnvVector::generateFunction4( vector<float> & vec, int repeat, int patternLength, int patternLengthFinal )
{
	int i, r;
	float l = (float)(patternLength-1);
	int lp1 = patternLength;
	int d1 = (lp1-1)/4;
	int d2 = (3*(lp1-1))/4;

	for (r = 0; r <= repeat; ++r)
	{
		if (r == repeat)
		{
			l = (float)(patternLengthFinal-1);
			lp1 = patternLengthFinal;
			d1 = (lp1-1)/4;
			d2 = (3*(lp1-1))/4;			
		}
		
		for (i = 0; i < d1; ++i)
			vec[r*patternLength+i] = mybounded( 0.5f - ((float)(2*i))/l );

		for (; i < d2; ++i)
			vec[r*patternLength+i] = mybounded( ((float)(2*i))/l - 0.5f );

		for (; i < lp1; ++i)
			vec[r*patternLength+i] = mybounded( 2.5f - ((float)(2*i))/l );
	}
}


void EnvVector::generateFunction5( vector<float> & vec, int repeat, int patternLength, int patternLengthFinal )
{
	int i, r;
	float l = (float)(patternLength-1);
	int lp1 = patternLength;
	int d = (lp1-1)/2;

	for (r = 0; r <= repeat; ++r)
	{
		if (r == repeat)
		{
			l = (float)(patternLengthFinal-1);
			lp1 = patternLengthFinal;
			d = (lp1-1)/2;
		}
		
		for (i = 0; i < d; ++i)
			vec[r*patternLength+i] = mybounded( ((float)(2*i))/l );

		for (; i < lp1; ++i)
			vec[r*patternLength+i] = mybounded( ((float)(2*i))/l - 1.0f );
	}
}


void EnvVector::generateFunction6( vector<float> & vec, int repeat, int patternLength, int patternLengthFinal )
{
	int i, r;
	float l = (float)(patternLength-1);
	int lp1 = patternLength;
	int d1 = (lp1-1)/3;
	int d2 = (2*(lp1-1))/3;	

	for (r = 0; r <= repeat; ++r)
	{
		if (r == repeat)
		{
			l = (float)(patternLengthFinal-1);
			lp1 = patternLengthFinal;
			d1 = (lp1-1)/3;
			d2 = (2*(lp1-1))/3;			
		}
		
		for (i = 0; i < d1; ++i)
			vec[r*patternLength+i] = mybounded( ((float)(3*i))/(2.0f*l) );

		for (; i < d2; ++i)
			vec[r*patternLength+i] = 0.5f;

		for (; i < lp1; ++i)
			vec[r*patternLength+i] = mybounded( ((float)(3*i))/(2.0f*l) - 0.5f );
	}
}


void EnvVector::generateFunction7( vector<float> & vec, int repeat, int patternLength, int patternLengthFinal )
{
	int i, r;
	float l = (float)(patternLength-1);
	int lp1 = patternLength;
	int d = (lp1-1)/2;	

	for (r = 0; r <= repeat; ++r)
	{
		if (r == repeat)
		{
			l = (float)(patternLengthFinal-1);
			lp1 = patternLengthFinal;
			d = (lp1-1)/2;
		}
		
		for (i = 0; i < d; ++i)
			vec[r*patternLength+i] = mybounded( 0.5f - ((float)i)/l );

		for (; i < lp1; ++i)
			vec[r*patternLength+i] = mybounded( ((float)i)/l );
	}
}


void EnvVector::operatorUnion( int startPos, EnvVector & modifier )
{
	// fuzzy union (= OR, max)
	
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = mymax( m_Values[idx], vec[i] );
	}
}


void EnvVector::operatorIntersection( int startPos, EnvVector & modifier )
{
	// fuzzy intersection (= AND, min)
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = mymin( m_Values[idx], vec[i] );
	}
}


void EnvVector::operatorAvg( int startPos, EnvVector & modifier )
{
	// average
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = (m_Values[idx] + vec[i]) / 2.0f;
	}
}


void EnvVector::operatorCompAvg( int startPos, EnvVector & modifier )
{
	// Complement of average
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = 1.0f - ((m_Values[idx] + vec[i]) / 2.0f);
	}
}


void EnvVector::operatorCompIntersection( int startPos, EnvVector & modifier )
{
	// Complement of intersection
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = 1.0f - min( m_Values[idx], vec[i] );
	}
}


void EnvVector::operatorCompUnion( int startPos, EnvVector & modifier )
{
	// Complement of union
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = 1.0f - max( m_Values[idx], vec[i] );
	}
}


void EnvVector::operatorMult( int startPos, EnvVector & modifier )
{
	// A.B
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = m_Values[idx] * vec[i];
	}	
}


void EnvVector::operatorCompMult( int startPos, EnvVector & modifier )
{
	// Complement of (A.B)
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = 1.0f - (m_Values[idx] * vec[i]);
	}	
}


void EnvVector::operatorAUnionCompB( int startPos, EnvVector & modifier )
{
	// A union (Complement of B)
	
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = mymax( m_Values[idx], 1.0f - vec[i] );
	}
}


void EnvVector::operatorAIntersectionCompB( int startPos, EnvVector & modifier )
{
	// A intersection (Complement of B)
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = mymin( m_Values[idx], 1.0f - vec[i] );
	}
}


void EnvVector::operatorCompAUnionB( int startPos, EnvVector & modifier )
{
	// (Complement of A) union B
	
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = mymax( 1.0f - m_Values[idx], vec[i] );
	}
}


void EnvVector::operatorCompAIntersectionB( int startPos, EnvVector & modifier )
{
	// (Complement of A) intersection B
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = mymin( 1.0f - m_Values[idx], vec[i] );
	}
}


void EnvVector::operatorCompAUnionMult( int startPos, EnvVector & modifier )
{
	// (Complement of A) union A.B
	
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = mymax( 1.0f - m_Values[idx], m_Values[idx] * vec[i] );
	}
}


void EnvVector::operatorCompAIntersectionMult( int startPos, EnvVector & modifier )
{
	// (Complement of A) intersection A.B
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = mymin( 1.0f - m_Values[idx], m_Values[idx] * vec[i] );
	}
}


void EnvVector::operatorCompBUnionMult( int startPos, EnvVector & modifier )
{
	// (Complement of B) union A.B
	
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = mymax( 1.0f - vec[i], m_Values[idx] * vec[i] );
	}
}


void EnvVector::operatorCompBIntersectionMult( int startPos, EnvVector & modifier )
{
	// (Complement of B) intersection A.B
	vector<float> & vec = modifier.getValues();
	int l1 = m_Values.size();
	int l2 = vec.size();
	int idx;
	for (int i = 0; i < l2; ++i)
	{
		idx = (startPos+i) % l1;
		m_Values[idx] = mymin( 1.0f - vec[i], m_Values[idx] * vec[i] );
	}
}


ostream& operator << (ostream& os, const EnvVector& id)
{
	vector<float>::const_iterator env_it  = id.m_Values.begin();
	vector<float>::const_iterator env_end = id.m_Values.end();
	//int oldval = os.width(5);
	os << id.m_Values.size() << " ";
	while (env_it != env_end)
	{
		os << (*env_it) << " ";
		++env_it;
	}
	//os.width(oldval);
	return os;
}
