#include "NRCRandom.h"
#include <math.h>
#define PI 3.141592654

double bnldev(double pp, int n, long *idum)
{
        int j;
        static int nold=(-1);
        double am,em,g,angle,p,bnl,sq,t,y;
        static double pold=(-1.0),pc,plog,pclog,en,oldg;

        p=(pp <= 0.5 ? pp : 1.0-pp);
        am=n*p;
        if (n < 25) {
                bnl=0.0;
                for (j=1;j<=n;j++)
                        if (ran1(idum) < p) ++bnl;
        } else if (am < 1.0) {
                g=exp(-am);
                t=1.0;
                for (j=0;j<=n;j++) {
                        t *= ran1(idum);
                        if (t < g) break;
                }
                bnl=(j <= n ? j : n);
        } else {
                if (n != nold) {
                        en=n;
                        oldg=gammln(en+1.0);
                        nold=n;
                } if (p != pold) {
                        pc=1.0-p;
                        plog=log(p);
                        pclog=log(pc);
                        pold=p;
                }
                sq=sqrt(2.0*am*pc);
                do {
                        do {
                                angle=PI*ran1(idum);
                                y=tan(angle);
                                em=sq*y+am;
                        } while (em < 0.0 || em >= (en+1.0));
                        em=floor(em);
                        t=1.2*sq*(1.0+y*y)*exp(oldg-gammln(em+1.0)
                                -gammln(en-em+1.0)+em*plog+(en-em)*pclog);
                } while (ran1(idum) > t);
                bnl=em;
		}
        if (p != pp) bnl=n-bnl;
        return bnl;
}
#undef PI

/* (C) Copr. 1986-92 Numerical Recipes Software D0>4^c1&]v%]w#o+"#. */

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

double ran1(long *idum)
{
        int j;
        long k;
        static long iy=0;
        static long iv[NTAB];
        double temp;

        if (*idum <= 0 || !iy) {
                if (-(*idum) < 1) *idum=1;
                else *idum = -(*idum);
                for (j=NTAB+7;j>=0;j--) {
                        k=(*idum)/IQ;
                        *idum=IA*(*idum-k*IQ)-IR*k;
                        if (*idum < 0) *idum += IM;
                        if (j < NTAB) iv[j] = *idum;
                }
                iy=iv[0];
        }
        k=(*idum)/IQ;
        *idum=IA*(*idum-k*IQ)-IR*k;
        if (*idum < 0) *idum += IM;
        j=iy/NDIV;
        iy=iv[j];
        iv[j] = *idum;
        if ((temp=AM*iy) > RNMX) return RNMX;
        else return temp;
}
#undef IA
#undef IM
#undef AM
#undef IQ
#undef IR
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX

/* (C) Copr. 1986-92 Numerical Recipes Software D0>4^c1&]v%]w#o+"#. */

double gammln(double xx)
{
        double x,y,tmp,ser;
        static double cof[6]={76.18009172947146,-86.50532032941677,
                24.01409824083091,-1.231739572450155,
                0.1208650973866179e-2,-0.5395239384953e-5};
        int j;

        y=x=xx;
        tmp=x+5.5;
        tmp -= (x+0.5)*log(tmp);
        ser=1.000000000190015;
        for (j=0;j<=5;j++) ser += cof[j]/++y;
        return -tmp+log(2.5066282746310005*ser/x);
}
/* (C) Copr. 1986-92 Numerical Recipes Software D0>4^c1&]v%]w#o+"#. */


double gasdev(long *idum)
/* Returns a normally distributed deviate with zero mean and unit variance, 
using ran1(idum) as the source of uniform deviates. */
{
  static int iset=0;
  static double gset;
  double fac, rsq, v1, v2;

  if (iset == 0) {
    do {
      v1=2.0*ran1(idum)-1.0;
      v2=2.0*ran1(idum)-1.0;
      rsq=v1*v1+v2*v2;
    } while (rsq >= 1.0 || rsq == 0.0);
    fac=sqrt(-2.0*log(rsq)/rsq);

    gset=v1*fac;
    iset=1;
    return v2*fac;
  } else {
    iset=0;
    return gset;
  }
}
